/////////////////////////////////////////////////////////////////////////
//                                                                     //
// This file contains the declaration of the stereogram generation     //
// function.                                                           //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
#ifndef __STEREOGRAM_HPP__
#define __STEREOGRAM_HPP__


class CScene;
class CImage;


void GenerateStereogram(const CScene* pScene,
                        const CImage& depthImage,
                        CImage& stereogram);

void MapDepthTo0_1(const CScene* pScene,
                   CImage& depthImage);


#endif // __STEREOGRAM_HPP__
